#include "ChatMsgThread.h"
#include "MySyncObject.h"
#include "SendRecvThread.h"
#include "RingBuff.h"



CChatMsgThread::CChatMsgThread(CSendRecvThread *pCSendRecvThread)
{
	// パラメータを記憶する
	m_pCSendRecvThread = pCSendRecvThread;

	// 変数の初期化
	m_fIamZombie = FALSE;

	m_pCRingBuffCmd = new CRingBuff(CMD_BUFF_SIZE);					// コマンド格納用
	m_pCRingBuffForDistributeChat = new CRingBuff(RECV_BUFF_SIZE);	// チャット配信用リングバッファの構築
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
}


CChatMsgThread::~CChatMsgThread()
{
	if (m_pCMySyncObject != NULL)
		m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	SAFE_DELETE(m_pCRingBuffCmd)
	SAFE_DELETE(m_pCRingBuffForDistributeChat)		// チャット配信用リングバッファの破棄
}


UINT CChatMsgThread::DoWork()
{
	BOOL		fRet = TRUE;

	while (!m_fStopFlag)
	{
		usleep(10 * 1000);			// usleepを入れて他のスレッドにチャンスを与える
		if (AnalyzeCmd() == -1)		// 返値が-1の時がエラー、このスレッドをゾンビにする
		{
			fRet = FALSE;
			break;
		}
	}
	m_pCMySyncObject->Lock();
	m_fIamZombie = TRUE;
	m_pCMySyncObject->UnLock();
	return((fRet == TRUE) ? 0 : -1);
}

//----------------------------------------------
// function
//		メッセージコマンドの解析
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗
//----------------------------------------------
int CChatMsgThread::AnalyzeCmd()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	int			iSize = m_pCRingBuffCmd->GetReadableSize();
	int			iPacketSize, iDataSize;
	LPBYTE		pbPacket = NULL;

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする(パケット丸ごと格納されているはずですが)
	m_pCRingBuffCmd->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	wCmd = ntohs(Header.wCommand);
	iDataSize = ntohs(Header.wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);
	if (iSize < iPacketSize)				// パケットサイズに満たないときは何もしない
		goto L_END;

	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を読む
	m_pCRingBuffCmd->Read(pbPacket, iPacketSize);

	switch (wCmd)
	{
		case CMD_MSG_DATA:
			iRet = RecvMessagePacket((MsgDataRec *)pbPacket);
			break;
		default:
			iRet = -1;							// 知らないコマンド
			break;
	}
L_END:
	SAFE_FREE(pbPacket)
	return(iRet);
}

//----------------------------------------------
// function
//		メッセージコマンドの受信
// parameter
//		 MsgDataRec *pbPacket	[in]メッセージパケット
// return
//		0:パケットが揃っていない
//		1:パケットを受信したので処理を行った
//		-1:エラーが発生した
//----------------------------------------------
int CChatMsgThread::RecvMessagePacket(MsgDataRec *pbPacket)
{
	int			iRet = 0;
	int			iMsgSize, iSize;
	char		*pszMsg = NULL;

	iMsgSize = ntohs(pbPacket->header.wDataLen);
	// NULLターミネート分を追加して確保
	pszMsg = (char *)calloc(iMsgSize + 1, sizeof(char));
	memcpy(pszMsg, pbPacket->bMsgData, iMsgSize);

	// チャット送信のためにメッセージデータをパケット化してセット
	SendMessagePacket(pszMsg, iMsgSize);

	Locate(12, 10, 1);
	fprintf(stderr, "Msg:recv %s\n", pszMsg);

	SAFE_FREE(pszMsg)
	iRet = 1;
	return(iRet);
}

//----------------------------------------------
// function
//		メッセージパケットの送信
//		チャットメッセージをクライアントに転送するためのエリアに
//		格納する
// parameter
//		char	*pcData		[in]UTF-8データ(NULLターミネートなし)
//		int		iSize		[in]データサイズ
// retun
//		なし
//----------------------------------------------
BOOL CChatMsgThread::SendMessagePacket(char *pcData, int iSize)
{
	MsgDataRec	*pMsgData = NULL;
	int			iPacketSize = sizeof(HeaderRec) + iSize;		// パケットサイズ
	BOOL		fRet = FALSE;

	pMsgData = (MsgDataRec *)calloc(iPacketSize, sizeof(BYTE));	// パケット全体のエリアを確保

	memcpy(pMsgData->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pMsgData->header.wCommand = htons(CMD_MSG_DATA);
	pMsgData->header.wDataLen = htons(iSize);

	memcpy(pMsgData->bMsgData, pcData, iSize);
	// チャット送信のためにメッセージパケットデータをセット
	fRet = SetDataForDistributeChat((char *)pMsgData, iPacketSize);

	SAFE_FREE(pMsgData)
	return(fRet);
}

//----------------------------------------------
// function
//		チャットメッセージをクライアントに分配するために格納する
// parameter
//		char	*pcData	[in]チャットメッセージパケット
//		int		iSize	[in]パケット長
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CChatMsgThread::SetDataForDistributeChat(char *pcData, int iSize)
{
	BOOL	fRet = FALSE;

	// 格納用リングバッファに空きがないときは書き込まない
	fRet = m_pCRingBuffForDistributeChat->Write((LPBYTE)pcData, iSize);
	return(fRet);
}


//----------------------------------------------
// function
//		チャット分配用データの取得
//		パケット単位で取り出す
//		今回は１パケット分
// parameter
//		char **ppcData	[in/out]チャット転送用データ
// return
//		データ長
//----------------------------------------------
int CChatMsgThread::GetDataForDistributeChat(char **ppcData)
{
	int			iSize = 0, iStoreSize, iPacketSize;
	HeaderRec	Header;

	if ((iStoreSize = m_pCRingBuffForDistributeChat->GetReadableSize()) >= sizeof(HeaderRec))
	{
		// ヘッダ部を仮り読み込みする
		m_pCRingBuffForDistributeChat->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
		iPacketSize = ntohs(Header.wDataLen) + sizeof(HeaderRec);
		if (iStoreSize >= iPacketSize)
		{
			*ppcData = (char *)calloc(iPacketSize, sizeof(char));
			iSize = m_pCRingBuffForDistributeChat->Read((LPBYTE)*ppcData, iPacketSize);
		}
	}
	return(iSize);
}


//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
BOOL CChatMsgThread::IsZombie()
{
	BOOL	fRet;

	m_pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pCMySyncObject->UnLock();
	return(fRet);
}

