#include "RecvFileThread.h"
#include "MySyncObject.h"
#include "SendRecvThread.h"
#include "RingBuff.h"

CRecvFileThread::CRecvFileThread(CSendRecvThread *pCSendRecvThread)
{
	// パラメータを記憶する
	m_pCSendRecvThread = pCSendRecvThread;

	// 変数の初期化
	m_fIamZombie = FALSE;
	m_fdDstFile = -1;							// ファイルディスクリプタ
	m_iRecvFileStatus = RECV_FILE_STAT_IDLE;	// ファイル送信コマンド処理の遷移
	memset(m_szFileName, 0, sizeof(m_szFileName)); // ファイル削除のためのファイル名の初期化

	m_pCRingBuffCmd = new CRingBuff(CMD_BUFF_SIZE);	// ファイル送信コマンド格納用
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
}


CRecvFileThread::~CRecvFileThread()
{
	if (m_pCMySyncObject != NULL)
		m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	SAFE_DELETE(m_pCRingBuffCmd)
// オープンしたままのファイルがあればおかしいのでクローズ、削除
	if (m_fdDstFile != -1)
	{
		close(m_fdDstFile);
		unlink(m_szFileName);
	}
	m_fdDstFile = -1;
}

UINT CRecvFileThread::DoWork()
{
	BOOL		fRet = TRUE;

	while (!m_fStopFlag)
	{
		usleep(10 * 1000);			// usleepを入れて他のスレッドにチャンスを与える
		if (AnalyzeCmd() == -1)		// 返値が-1の時がエラー、このスレッドをゾンビにする
		{
			fRet = FALSE;
			break;
		}
	}
	m_pCMySyncObject->Lock();
	m_fIamZombie = TRUE;
	m_pCMySyncObject->UnLock();
	return((fRet == TRUE) ? 0 : -1);
}

//----------------------------------------------
// function
//		ファイル送信コマンドの解析
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗
//----------------------------------------------
int CRecvFileThread::AnalyzeCmd()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	int			iSize = m_pCRingBuffCmd->GetReadableSize();
	int			iPacketSize, iDataSize;
	LPBYTE		pbPacket = NULL;

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする(パケット丸ごと格納されているはずですが)
	m_pCRingBuffCmd->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	wCmd = ntohs(Header.wCommand);
	iDataSize = ntohs(Header.wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);
	if (iSize < iPacketSize)				// パケットサイズに満たないときは何もしない
		goto L_END;

	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を読む
	m_pCRingBuffCmd->Read(pbPacket, iPacketSize);

	switch (wCmd)
	{
		case CMD_SEND_FILE_START_REQ:			// ①送信開始要求受信
			iRet = RecvSendFileStartPacket((SendFileStartReqRec *)pbPacket);
			break;
		case CMD_SEND_FILE_ING:					// ②ファイル送信中
			iRet = RecvSendFileIngPacket((SendFileIngRec *)pbPacket);
			break;
		case CMD_SEND_FILE_END_REQ:				// ③送信完了要求受信
			iRet = RecvSendFileEndPacket((SendFileEndReqRec *)pbPacket);
			break;
		case CMD_SEND_FILE_ABORT_REQ:			// 送信中断要求受信
			iRet = RecvSendFileAbortPacket((SendFileAbortReqRec *)pbPacket);
			break;
		default:
			iRet = -1;							// 知らないコマンド
			break;
	}
L_END:
	SAFE_FREE(pbPacket)
	return(iRet);
}

//----------------------------------------------
// function
//		①ファイル送信開始コマンドの受信
// parameter
//		 SendFileStartReqRec *pbPacket	[in]ファイル送信開始コマンドパケット
// return
//		1:解析成功 -1:致命的エラー
//----------------------------------------------
int CRecvFileThread::RecvSendFileStartPacket(SendFileStartReqRec *pbPacket)
{
	int			iRet = 0;

	Locate(1, 11, 2);
	fprintf(stderr, "ファイル受信 %s %s(%u)", pbPacket->FileInfo.szFileName, pbPacket->FileInfo.szFileSize, timeGetTime());

	if (m_iRecvFileStatus != RECV_FILE_STAT_IDLE)
	{
		iRet = -1;						// 遷移状態エラーは切断する
		goto L_END;
	}
	// ここで書き込み用にファイルをopenする
	if ((m_fdDstFile = open(pbPacket->FileInfo.szFileName,
		O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)) == -1)
	{
		SendSendFileStartResPacket(-1);		// エラー発生を送信
		goto L_END;
	}
	// 開始したファイル名を覚える
	DISABLE_C4996
	strcpy(m_szFileName, pbPacket->FileInfo.szFileName);
	ENABLE_C4996
	m_iRecvFileStatus = RECV_FILE_STAT_ING;
	m_pCRingBuffCmd->Clear();				// ファイル送信コマンドバッファを空にする
	SendSendFileStartResPacket(1);			// 準備OKを送信
	iRet = 1;
L_END:
	if (iRet == -1)					// エラー時オープンしているファイルがあればクローズ、削除
	{
		if (m_fdDstFile != -1)
		{
			close(m_fdDstFile);
			unlink(m_szFileName);
		}
		m_fdDstFile = -1;
	}
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信開始要求に対する応答を送信する
// parameter
//		char cResCode	[in]応答コード 1/-1
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendSendFileStartResPacket(char cResCode)
{
	BOOL	fRet = FALSE;
	SendFileResRec	*pSendFileStartRes = NULL;
	int		iPacketSize;

	iPacketSize = sizeof(SendFileResRec);
	pSendFileStartRes = (SendFileResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pSendFileStartRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileStartRes->header.wCommand = htons(CMD_SEND_FILE_START_RES);
	pSendFileStartRes->header.wDataLen = htons(sizeof(char));

	pSendFileStartRes->cResCode = cResCode;

	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileStartRes, iPacketSize);
	SAFE_FREE(pSendFileStartRes)


	return(fRet);
}

//----------------------------------------------
// function
//		ファイル送信中コマンドの受信
// parameter
//		 SendFileIngRec *pbPacket[in]ファイル送信中コマンドパケット
// return
//		0:十分なデータが格納されていない 1:解析成功 -1:解析失敗
//----------------------------------------------
int CRecvFileThread::RecvSendFileIngPacket(SendFileIngRec *pbPacket)
{
	int			iRet = 0;
	long		lDataSize, lOffset;

	Locate(1, 12, 1);
	fprintf(stderr, "ファイル受信中 %s %s", pbPacket->Stat.szOffset, pbPacket->Stat.szSize);

	if (m_iRecvFileStatus != RECV_FILE_STAT_ING)
	{
		iRet = -1;				// 遷移状態エラーは切断する
		goto L_END;
	}
	lDataSize = atol(pbPacket->Stat.szSize);
	lOffset = atol(pbPacket->Stat.szOffset);
	if (lDataSize > 0)
	{
		// ファイル操作失敗時エラーとして切断
		if (lseek(m_fdDstFile, lOffset, SEEK_SET) == -1)
		{
			iRet = -1;
			goto L_END;
		}
		if (write(m_fdDstFile, pbPacket->bData, lDataSize) != lDataSize)
		{
			iRet = -1;
			goto L_END;
		}
	}
	iRet = 1;
L_END:
	if (iRet == -1)				// エラー時オープンしているファイルがあればクローズ、削除
	{
		if (m_fdDstFile != -1)
		{
			close(m_fdDstFile);
			unlink(m_szFileName);
		}
		m_fdDstFile = -1;
	}
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信終了コマンドの受信
// parameter
//		 SendFileEndReqRec *pbPacket 	[in]ファイル送信終了コマンドパケット
// return
//		0:十分なデータが格納されていない 1:解析成功 -1:解析失敗
//----------------------------------------------
int CRecvFileThread::RecvSendFileEndPacket(SendFileEndReqRec *pbPacket)
{
	int			iRet = 0;
	long		lDataSize, lOffset;

	Locate(1, 13, 1);
	fprintf(stderr, "ファイル受信完了 %s %s(%u)", pbPacket->Stat.szOffset, pbPacket->Stat.szSize, timeGetTime());

	if (m_iRecvFileStatus != RECV_FILE_STAT_ING)
	{
		iRet = -1;				// 遷移状態エラーは切断する
		goto L_END;
	}
	lDataSize = atol(pbPacket->Stat.szSize);
	lOffset = atol(pbPacket->Stat.szOffset);
	if (lDataSize > 0)
	{
		// ファイル操作失敗時エラーとして切断
		if (lseek(m_fdDstFile, lOffset, SEEK_SET) == -1)
		{
			iRet = -1;
			goto L_END;
		}
		if (write(m_fdDstFile, pbPacket->bData, lDataSize) != lDataSize)
		{
			iRet = -1;
			goto L_END;
		}
	}
	iRet = 1;
	SendSendFileEndResPacket(1);		// 送信完了要求応答送信
L_END:
	if (m_fdDstFile != -1)
	{
		close(m_fdDstFile);
		if(iRet == -1)					// エラー時削除
			unlink(m_szFileName);
	}
	m_iRecvFileStatus = RECV_FILE_STAT_IDLE;
	m_fdDstFile = -1;
	return(iRet);
}


//----------------------------------------------
// function
//		ファイル送信完了要求に対する応答を送信する
// parameter
//		char cResCode	[in]応答コード 1/-1
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendSendFileEndResPacket(char cResCode)
{
	BOOL	fRet = FALSE;
	SendFileResRec	*pSendFileEndRes = NULL;
	int		iPacketSize;

	iPacketSize = sizeof(SendFileResRec);
	pSendFileEndRes = (SendFileResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pSendFileEndRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileEndRes->header.wCommand = htons(CMD_SEND_FILE_END_RES);
	pSendFileEndRes->header.wDataLen = htons(sizeof(char));

	pSendFileEndRes->cResCode = cResCode;

	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileEndRes, iPacketSize);
	SAFE_FREE(pSendFileEndRes)

	return(fRet);
}

//----------------------------------------------
// function
//		ファイル送信中断コマンドの受信
// parameter
//		 SendFileAbortReqRec *pbPacket 	[in]ファイル送信中断コマンドパケット
// return
//		0:十分なデータが格納されていない 1:解析成功 -1:解析失敗
//----------------------------------------------
int CRecvFileThread::RecvSendFileAbortPacket(SendFileAbortReqRec *pbPacket)
{
	int			iRet = 0;

	m_iRecvFileStatus = RECV_FILE_STAT_ABORT;
	Locate(1, 13, 1);
	fprintf(stderr, "ファイル受信中断");

	iRet = 1;
	SendSendFileAbortResPacket(1);		// 送信中断要求応答送信

	if (m_fdDstFile != -1)
	{
		close(m_fdDstFile);
		unlink(m_szFileName);
	}
	m_fdDstFile = -1;
	m_iRecvFileStatus = RECV_FILE_STAT_IDLE;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル送信中断要求に対する応答を送信する
// parameter
//		char cResCode	[in]応答コード 1/-1
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CRecvFileThread::SendSendFileAbortResPacket(char cResCode)
{
	BOOL	fRet = FALSE;
	SendFileResRec	*pSendFileAbortRes = NULL;
	int		iPacketSize;

	iPacketSize = sizeof(SendFileResRec);
	pSendFileAbortRes = (SendFileResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pSendFileAbortRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pSendFileAbortRes->header.wCommand = htons(CMD_SEND_FILE_ABORT_RES);
	pSendFileAbortRes->header.wDataLen = htons(sizeof(char));
	pSendFileAbortRes->cResCode = cResCode;

	fRet = m_pCSendRecvThread->SetSendData((char *)pSendFileAbortRes, iPacketSize);
	SAFE_FREE(pSendFileAbortRes)
	return(fRet);
}

//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
BOOL CRecvFileThread::IsZombie()
{
	BOOL	fRet;

	m_pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pCMySyncObject->UnLock();
	return(fRet);
}
