#include "SendFileThread.h"
#include "MySyncObject.h"
#include "SendRecvThread.h"
#include "RingBuff.h"

CSendFileThread::CSendFileThread(CSendRecvThread *pCSendRecvThread)
{
	// パラメータを記憶する
	m_pCSendRecvThread = pCSendRecvThread;

	// 変数の初期化
	m_fIamZombie = FALSE;
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
	m_pCRingBuffCmd = new CRingBuff(CMD_BUFF_SIZE);	// ファイル受信コマンド格納用

	m_iSendFileStatus = SEND_FILE_STAT_IDLE;		// ファイル送信処理の遷移

	m_cRecvFileStartReq = 0;						// ファイル送信開始要求(0:初期値 1:OK -1:エラー)
	m_cRecvFileEndRes = 0;							// ファイル送信完了要求応答(0:初期値 1:OK -1:エラー)
	m_dwStart = 0;									// 応答待ちタイムアウトのため要求を行った時刻
	m_fdSrcFile = -1;								// 送信ファイルディスクリプタ
	m_lFileLength = 0;								// 送信ファイルの長さ
	m_lTotalSentSize = 0;							// 送信済みバイト数
	m_iWaitCount = 0;								// 送信待ちをした回数
}


CSendFileThread::~CSendFileThread()
{
	if (m_pCMySyncObject != NULL)
		m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	SAFE_DELETE(m_pCRingBuffCmd)					// ファイル受信要求/応答格納用
}

UINT CSendFileThread::DoWork()
{
	int			iRet = 0;

	while (!m_fStopFlag)
	{
		usleep(5 * 1000);
		// ファイル送信遷移状態に応じた処理の実行
		// 応答に対する処理とファイル送信処理をこの関数で実行します
		if((iRet = DoSendFileState()) == -1)
			break;
	}
	if (m_fdSrcFile != -1)
		close(m_fdSrcFile);
	m_fdSrcFile = -1;
	m_pCMySyncObject->Lock();
	m_fIamZombie = TRUE;		// ファイル送信失敗でゾンビ
	m_pCMySyncObject->UnLock();
	return((iRet != -1) ? 0 : -1);
}

//----------------------------------------------
// function
//		ファイル送信遷移状態に応じた処理を実行
// parameter
//		なし
// return
//		0:何もしなかった 1:処理を実行した 2:送信完了 -1:エラー発生
//----------------------------------------------
int CSendFileThread::DoSendFileState()
{
	int iRet = 0;

	if (AnalyzeCmd() == -1)						// 要求・応答などの解析 返値が-1の時がエラー
	{
		iRet = -1;
		goto L_END;
	}
	switch (m_iSendFileStatus)
	{
		case SEND_FILE_STAT_START:				// ファイル受信開始要求待ち->送信中/初期状態
			if ((iRet = CheckRecvFileStartReq(TIMEOUT_FOR_RES)) == 1)
			{
				if (m_lFileLength == 0)			// ファイル受信を実施しない要求
					m_iSendFileStatus = SEND_FILE_STAT_IDLE;
				else
					m_iSendFileStatus = SEND_FILE_STAT_ING;
			}
			break;
		case SEND_FILE_STAT_ING:				// 送信中->送信中/受信完了応答待ち
			if ((iRet = SendFileIng()) == 2)	// 送信中->受信完了応答待ち
			{
				iRet = 1;						// 完了要求応答待ちまで完了としない
				m_dwStart = timeGetTime();
				m_iSendFileStatus = SEND_FILE_STAT_END;
			}
			break;
		case SEND_FILE_STAT_END:				// 受信完了要求応答待ち->初期状態
			if ((iRet = CheckRecvFileEndRes(TIMEOUT_FOR_RES)) == 1)
				m_iSendFileStatus = SEND_FILE_STAT_IDLE;
			break;
		default:
			break;
	}
L_END:
	if (iRet == -1)
	{
		Locate(1, 11, 1);
		fprintf(stderr, "ファイル送信エラー");
	}
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信開始要求のチェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 1:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int CSendFileThread::CheckRecvFileStartReq(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cRecvFileStartReq == -1)				// エラー応答
		iRet = -1;
	else if (m_cRecvFileStartReq == 1)			// 成功応答
		iRet = 1;
L_END:
	return(iRet);
}

#define	READ_SIZE	(1024 * 10)			// ファイルの読み込み単位(ガバッと読みます)
//----------------------------------------------
// function
//		ファイルを読み込んでファイル送信中/ファイル受信完了パケットを送信する
//		この関数でエラーが起こったときは、中断コマンドを送るのも良いでしょう
// parameter
//		なし
// retun
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::SendFileIng()
{
	BOOL			fEnd = FALSE, fRet;
	LPBYTE			pbData = NULL;
	int				iRet = 0, iReadSize;
	RecvFileIngRec	*pRecvFileIng = NULL;
	int				iPacketSize;

	// ファイルを読み込む
	pbData = (LPBYTE)calloc(READ_SIZE, sizeof(BYTE));
	lseek(m_fdSrcFile, m_lTotalSentSize, SEEK_SET);
	iReadSize = read(m_fdSrcFile, pbData, READ_SIZE);
	iPacketSize = sizeof(HeaderRec) + sizeof(SendFileStatRec) + iReadSize;
	pRecvFileIng = (RecvFileIngRec *)calloc(iPacketSize, sizeof(BYTE));

	// 共通ヘッダ部にデータを格納する
	memcpy(pRecvFileIng->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileIng->header.wDataLen = htons((WORD)sizeof(SendFileStatRec) + iReadSize);
	// データ部にデータを格納する
	DISABLE_C4996
	sprintf(pRecvFileIng->Stat.szOffset, "%d", m_lTotalSentSize);	// 読み込み位置
	sprintf(pRecvFileIng->Stat.szSize, "%d", iReadSize);			// 読み込んだデータ数
	ENABLE_C4996
	if (iReadSize > 0)				// ファイルの途中を送信
	{
		Locate(1, 12, 1);
		fprintf(stderr, "ファイル送信中 %d %d %d", m_lTotalSentSize, iReadSize, m_lFileLength);
		// 共通ヘッダ部にデータを格納する
		pRecvFileIng->header.wCommand = htons(CMD_RECV_FILE_ING);
		// データ部にデータを格納する
		memcpy(pRecvFileIng->bData, pbData, iReadSize);
	}
	else							// ファイルの最後をまで読んだので完了を送信
	{
		Locate(1, 13, 1);
		fprintf(stderr, "ファイル送信完了 %d %d", m_lTotalSentSize, iReadSize);
		// 共通ヘッダ部にデータを格納する
		pRecvFileIng->header.wCommand = htons(CMD_RECV_FILE_END_REQ);
		fEnd = TRUE;				// ファイル送信完了
	}
	// パケット全体を送信バッファに書きこむ
	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileIng, iPacketSize);
	if (fRet == TRUE)				// パケットを書き込めた
	{
		if (fEnd == TRUE)			// 最後まで送信したので終わり
		{
			close(m_fdSrcFile);
			m_fdSrcFile = -1;
			iRet = 2;
		}
		else
		{
			m_lTotalSentSize += iReadSize;	// パケットを書き込めたので読み込みポインタ更新
			iRet = 1;
		}
	}
	else							// 書き込めなかったので読み込みポインタはそのまま
	{
		Locate(1, 14, 0);
		fprintf(stderr, "送信バッファの空き待ち(%d)", ++m_iWaitCount);
		// 送信バッファに書けなかったのでちょっと待つ
		usleep(100 * 1000);			// READ_SIZEを送信するのに掛かる時間程度待つのが良いでしょう
	}
	SAFE_FREE(pRecvFileIng)
	SAFE_FREE(pbData)
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信開完了要求応答のチェック
// parameter
//		DWORD	dwTimeOut	[in]タイムアウト
// return
//		0:まだ来ていない 1:成功が来た -1:失敗またはタイムアウト
//----------------------------------------------
int CSendFileThread::CheckRecvFileEndRes(DWORD dwTimeOut)
{
	int		iRet = 0;

	if (GetdwInterval(timeGetTime(), m_dwStart) >= dwTimeOut)	// タイムアウト
	{
		iRet = -1;
		goto L_END;
	}
	if (m_cRecvFileEndRes == -1)				// エラー応答
		iRet = -1;
	else if (m_cRecvFileEndRes == 1)			// 成功応答
		iRet = 1;
L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信コマンド/応答の解析
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗
//----------------------------------------------
int CSendFileThread::AnalyzeCmd()
{
	int			iRet = 0;
	HeaderRec	Header;
	WORD		wCmd;
	int			iSize = m_pCRingBuffCmd->GetReadableSize();
	int			iPacketSize, iDataSize;
	LPBYTE		pbPacket = NULL;

	if (iSize < sizeof(HeaderRec))			// ヘッダサイズに満たないときは何もしない
		goto L_END;

	// ヘッダ部を借り読み込みする(パケット丸ごと格納されているはず)
	m_pCRingBuffCmd->ReadWithoutUpdateHeadPoint((LPBYTE)&Header, sizeof(HeaderRec));
	wCmd = ntohs(Header.wCommand);
	iDataSize = ntohs(Header.wDataLen);
	iPacketSize = iDataSize + sizeof(HeaderRec);
	if (iSize < iPacketSize)				// パケットサイズに満たないときは何もしない
		goto L_END;

	pbPacket = (LPBYTE)calloc(iPacketSize, sizeof(BYTE));
	// パケット全体を読む
	m_pCRingBuffCmd->Read(pbPacket, iPacketSize);
	switch (wCmd)
	{
		case CMD_RECV_FILE_SIZE_REQ:			// 受信サイズ要求
			if ((iRet = RecvRecvFileSizeReqPacket((RecvFileSizeReqRec *)pbPacket)) == 1)
			{
				m_dwStart = timeGetTime();
				m_iSendFileStatus = SEND_FILE_STAT_START;
			}
			break;
		case CMD_RECV_FILE_START_REQ:			// 受信開始要求
			iRet = RecvRecvFileStartReqPacket((RecvFileStartReqRec *)pbPacket);
			break;
		case CMD_RECV_FILE_END_RES:				// 受信完了要求に対する応答
			iRet = RecvRecvFileEndResPacket((RecvFileResRec *)pbPacket);
			break;
		case CMD_RECV_FILE_ABORT_REQ:			// 中断要求
			iRet = RecvRecvFileAbortReqPacket((RecvFileAbortReqRec *)pbPacket);
			break;
		default:
			iRet = -1;
			break;
	}
L_END:
	SAFE_FREE(pbPacket)
	return(iRet);
}

//----------------------------------------------
// function
//		ファイルサイズ要求コマンドの解析
//		ファイルのオープン
//		応答の送信(オープン失敗時はサイズ項目を0で応答する
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗 2:送信ファイルがない
//----------------------------------------------
int CSendFileThread::RecvRecvFileSizeReqPacket(RecvFileSizeReqRec *pbPacket)
{
	int			iRet = 1;
	struct stat statInfo;			// ファイルの情報のための変数

	if (m_iSendFileStatus != SEND_FILE_STAT_IDLE)
	{
		iRet = -1;						// 遷移状態エラーは切断する
		goto L_END;
	}
	m_cRecvFileStartReq = 0;						// ファイル送信開始要求(0:初期値 1:OK -1:エラー)
	m_cRecvFileEndRes = 0;							// ファイル送信完了要求応答(0:初期値 1:OK -1:エラー)
	m_lFileLength = 0;								// 送信ファイルの長さ
	m_lTotalSentSize = 0;							// 送信済みバイト数
	m_iWaitCount = 0;								// 送信待ちをした回数

	// ここで読み込み用にファイルをopenする
	if ((m_fdSrcFile = open(pbPacket->FileInfo.szFileName, O_RDONLY)) == -1)
	{
		iRet = 2;
		Locate(1, 11, 2);
		fprintf(stderr, "送信ファイルが存在しない %s", pbPacket->FileInfo.szFileName);
		if (SendRecvFileSizeResPacket(pbPacket->FileInfo.szFileName, 0) == FALSE)			// ファイルサイズ0で応答
			iRet = -1;
		goto L_END;
	}
	// ファイルの情報を取得(ファイルサイズ)
	if (fstat(m_fdSrcFile, &statInfo) != 0)
		goto L_END;
	m_lFileLength = statInfo.st_size;
	if (SendRecvFileSizeResPacket(pbPacket->FileInfo.szFileName, m_lFileLength) == FALSE)	// 応答
		iRet = -1;

L_END:
	return(iRet);
}


//----------------------------------------------
// function
//		ファイルサイズ要求応答送信
// parameter
//		LPSTR	pszFileName	[in]ファイル名
//		long	lFileLength	[in]ファイル長
// retun
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::SendRecvFileSizeResPacket(LPSTR pszFileName, long lFileLength)
{
	BOOL		fRet = FALSE;
	RecvFileSizeResRec	*pRecvFileSizeRes = NULL;
	int			iPacketSize;

	// 共通ヘッダ部にデータを格納
	iPacketSize = sizeof(RecvFileSizeResRec);
	pRecvFileSizeRes = (RecvFileSizeResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileSizeRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileSizeRes->header.wCommand = htons(CMD_RECV_FILE_SIZE_RES);
	pRecvFileSizeRes->header.wDataLen = htons(sizeof(FileInfoRec));

	Locate(1, 11, 2);
	fprintf(stderr, "ファイル送信 %s %d", pszFileName, lFileLength);
	// データ部にデータを格納(ファイルサイズ、ファイル名)
	sprintf(pRecvFileSizeRes->FileInfo.szFileName, "%s", pszFileName);
	sprintf(pRecvFileSizeRes->FileInfo.szFileSize, "%d", lFileLength);

	// ファイルサイズ要求応答パケットの送信
	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileSizeRes, iPacketSize);

	SAFE_FREE(pRecvFileSizeRes);
	return(fRet);
}

//----------------------------------------------
// function
//		ファイル受信開始要求受信処理
//		サイズが0の時はファイル受信が行われないのでクローズする
// parameter
//		なし
// retun
//		0:十分なデータが格納されていないなど 1:解析成功 -1:解析失敗
//----------------------------------------------
int CSendFileThread::RecvRecvFileStartReqPacket(RecvFileStartReqRec *pbPacket)
{
	int			iRet = 1;

	if (m_iSendFileStatus != SEND_FILE_STAT_START)
	{
		iRet = -1;						// 遷移状態エラーは切断する
		goto L_END;
	}
	m_cRecvFileStartReq = 1;
	if(atol(pbPacket->FileInfo.szFileSize) == 0)
		m_lFileLength = 0;
	if (m_lFileLength == 0)
		close(m_fdSrcFile);

L_END:
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信完了要求応答処理
// parameter
//		SendFileResRec *pbPacket	[in]受信パケット
// return
//		TRUE/FALSE
//----------------------------------------------
int CSendFileThread::RecvRecvFileEndResPacket(SendFileResRec *pbPacket)
{
	int			iRet = 1;

	m_cRecvFileEndRes = pbPacket->cResCode;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信中断コマンドの受信
// parameter
//		 RecvFileAbortReqRec *pbPacket 	[in]ファイル受信中断コマンドパケット
// return
//		0:十分なデータが格納されていない 1:解析成功 -1:解析失敗
//----------------------------------------------
int CSendFileThread::RecvRecvFileAbortReqPacket(RecvFileAbortReqRec *pbPacket)
{
	int			iRet = 0;

	m_iSendFileStatus = SEND_FILE_STAT_ABORT;
	Locate(1, 13, 1);
	fprintf(stderr, "ファイル送信中断");

	iRet = 1;
	SendRecvFileAbortResPacket(1);		// ★送信中断要求応答送信

	if (m_fdSrcFile != -1)
		close(m_fdSrcFile);
	m_fdSrcFile = -1;
	m_iSendFileStatus = SEND_FILE_STAT_IDLE;
	return(iRet);
}

//----------------------------------------------
// function
//		ファイル受信中断要求に対する応答を送信する
// parameter
//		char cResCode	[in]応答コード 1/-1
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CSendFileThread::SendRecvFileAbortResPacket(char cResCode)
{
	BOOL	fRet = FALSE;
	RecvFileResRec	*pRecvFileAbortRes = NULL;
	int		iPacketSize;

	iPacketSize = sizeof(RecvFileResRec);
	pRecvFileAbortRes = (RecvFileResRec *)calloc(iPacketSize, sizeof(BYTE));
	memcpy(pRecvFileAbortRes->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pRecvFileAbortRes->header.wCommand = htons(CMD_RECV_FILE_ABORT_RES);
	pRecvFileAbortRes->header.wDataLen = htons(sizeof(char));
	pRecvFileAbortRes->cResCode = cResCode;

	fRet = m_pCSendRecvThread->SetSendData((char *)pRecvFileAbortRes, iPacketSize);
	SAFE_FREE(pRecvFileAbortRes)

	return(fRet);
}

//----------------------------------------------
// function
//    このスレッドはゾンビ状態か
//	  別スレッドから参照される
// parameter
//    なし
// return
//    0:正常 -1:エラー発生
//----------------------------------------------
BOOL CSendFileThread::IsZombie()
{
	BOOL	fRet;

	m_pCMySyncObject->Lock();
	fRet = m_fIamZombie;
	m_pCMySyncObject->UnLock();
	return(fRet);
}
