#pragma once
#include "ThreadJob.h"
#include "define.h"					// パケット作成のため

#define	CMD_BUFF_SIZE	(1024 * 64)	// ファイル送信コマンド/応答格納リングバッファのサイズ

// ファイル送信遷移状態
#define SEND_FILE_STAT_IDLE		0	// 初期状態
#define SEND_FILE_STAT_START	1	// ファイル受信開始要求待ち
#define SEND_FILE_STAT_ING		2	// ファイル送信中
#define SEND_FILE_STAT_END		3	// ファイル送信完了要求応答待ち
#define SEND_FILE_STAT_ABORT	4	// ファイル送信中断処理中

class CMySyncObject;				// CMySyncObjectを使うため
class CSendRecvThread;				// CSendRecvThreadを使うため
class CRingBuff;					// CRingBuffを使うため

class CSendFileThread :
	public CThreadJob
{
public:
	CSendFileThread(CSendRecvThread *pCSendRecvThread);
	~CSendFileThread();

	CRingBuff	*m_pCRingBuffCmd;			// ファイル受信コマンド/応答格納リングバッファ

	UINT DoWork() override;
	BOOL IsZombie();						// このスレッドはゾンビ状態か

private:
	CSendRecvThread	*m_pCSendRecvThread;	// 送受信スレッド(親スレッド)
	CMySyncObject	*m_pCMySyncObject;
	BOOL			m_fIamZombie;			// ゾンビ状態か

	int				m_iSendFileStatus;		// 遷移状態 SEND_FILE_STAT_IDLE/SEND_FILE_STAT_START/SEND_FILE_STAT_ING/SEND_FILE_STAT_END

	char			m_cRecvFileStartReq;	// ファイル受信開始要求(0:初期値 1:OK -1:エラー)
	char			m_cRecvFileEndRes;		// ファイル受信完了要求応答(0:初期値 1:OK -1:エラー)
	DWORD			m_dwStart;				// 応答待ちタイムアウトのため要求を行った時刻

	int				m_fdSrcFile;			// 送信ファイルディスクリプタ
	long			m_lFileLength;			// 送信ファイルの長さ
	long			m_lTotalSentSize;		// 送信済みバイト数
	int				m_iWaitCount;			// 送信待ちをした回数

	int DoSendFileState();						// 遷移状態に応じたファイル送信処理,応答処理を実行
// ファイル送信処理
	int CheckRecvFileStartReq(DWORD dwTimeOut);	// ファイル受信開始要求チェック
	int CheckRecvFileEndRes(DWORD dwTimeOut);	// ファイル受信完了要求応答チェック
	int SendFileIng();							// ファイル途中処理、ファイル送信完了処理

// 応答処理
	int AnalyzeCmd();													// ファイル受信コマンド/応答パケットの解析
	int RecvRecvFileSizeReqPacket(RecvFileStartReqRec *pbPacket);		// ファイルサイズ要求処理
	BOOL SendRecvFileSizeResPacket(LPSTR pszFileName, long lFileLength);// ファイルサイズ要求応答送信
	int RecvRecvFileStartReqPacket(RecvFileStartReqRec *pbPacket);		// ファイル受信開始要求処理
	int RecvRecvFileEndResPacket(RecvFileResRec *pbPacket);				// ファイル受信完了要求応答処理
	int RecvRecvFileAbortReqPacket(RecvFileAbortReqRec *pbPacket);		// ファイル受信中断要求処理
	BOOL SendRecvFileAbortResPacket(char cResCode);						// ファイル受信中断要求応答送信
};

