#pragma once
#include "ThreadJob.h"
#include "define.h"			// RecvMessagePacketの引数のため

class CMySyncObject;		// このクラスの使用することを宣言
class CRingBuff;			// CRingBuffを使用するため
class CRecvFileThread;		// CRecvFileThreadを使用するため
class CSendFileThread;		// CSendFileThreadを使用するため
class CChatMsgThread;		// CChatMsgThreadを使用するため

#define	SEND_BUFF_SIZE		(1024 * 64)					// 送信リングバッファのサイズ
#if 1
#define	SENDBUFSIZE			(1024 + sizeof(HeaderRec))	// 一度に送信するサイズの最大値(共通ヘッダ分を加えてあります)
#else
#define	SENDBUFSIZE			(1024 * 4)					// TCPで送信速度を1Mbpsより上げたいときは断片化処理をTCPのスタックに任せる
														// スタック(send)に渡すサイズを増やす
#endif
#define	RECV_BUFF_SIZE		(1024 * 64)					// 受信リングバッファのサイズ
#define	RCVBUFSIZE			(1024 * 4)					// 一度に読む最大受信サイズ

typedef struct
{
	SOCKET			fdClient;			// 接続済みソケット(acceptの結果)
	CMySyncObject	*pCMySyncObject;	// 同期オブジェクト
	WORD			wUDPSrcPort;		// ★UDP送信元ポート/0
	sockaddr		*ai_addr;			// ★相手先アドレス
	size_t			ai_addrlen;			// ★相手先アドレス長
} ConnectionInfoRec;


class CSendRecvThread :
	public CThreadJob
{
public:
	CSendRecvThread(ConnectionInfoRec *pConInfo);	// パラメータをコンストラクタで渡す
	~CSendRecvThread();

	// 基底クラスの関数をオーバーライドする
	// C++11で明示的にoverrideを書くことが出来るようになりました
	// 基底クラスの当該関数にvirtualが書いていないとエラーを出してくれます
	UINT DoWork() override;							// DoRecvで実施している内容を記述
	// ★送信データの設定 コントロール用に強制的にTCPで送信する機能を追加
	BOOL SetSendData(char *pcData, int iSize, BOOL fForceTCP = FALSE);
	int GetDataForDistributeChat(char **ppcData);	// チャット分配用データの取得

	BOOL IsZombie();								// このスレッドはゾンビ状態か

private:
	ConnectionInfoRec *m_pConInfo;				// コンストラクタで渡されるパラメータを格納
												// このスレッド実行中領域が確保されていること
	BOOL m_fIamZombie;							// ゾンビ状態かどうかを保持
	CRingBuff *m_pCRingBuffSend;				// 送信データ格納用リングバッファ(TCP)
	CRingBuff *m_pCRingBuffRecv;				// 受信データ格納用リングバッファ(TCP)
	CRingBuff *m_pCRingBuffSendUDP;				// ★送信データ格納用リングバッファ(UDP)
	CRingBuff *m_pCRingBuffRecvUDP;				// ★受信データ格納用リングバッファ(UDP)

	sockaddr_storage m_PeerAddr;				// ★相手先情報(UDP送信用)
	socklen_t m_iPeerAddrLen;					// ★相手先情報の長さ
	SOCKET m_fdUDP;								// ★データ送受信方法がUDP使用時のソケット
	BYTE m_bTarnsMethod;						// ★0:TCP 1:UDP

	DWORD	m_dwPrevSentTime;					// 前回送信した時刻(msec)TCP
	DWORD	m_dwSendInterval;					// 次回送信までの間隔(msec)TCP
	DWORD	m_dwPrevSentTimeUDP;				// ★前回送信した時刻(msec)UDP
	DWORD	m_dwSendIntervalUDP;				// ★次回送信までの間隔(msec)UDP

	CRecvFileThread *m_pCRecvFileThread;		// ファイル受信スレッド
	CSendFileThread *m_pCSendFileThread;		// ファイル送信スレッド
	CChatMsgThread *m_pCChatMsgThread;			// チャットメッセージ処理スレッド

	int GetSendData(char **ppcData, int iTcpUdp);	// ★送信データの取得(TCP/UDPで参照バッファ切り替え)

	BOOL RecvTCPData(pollfd *fds);				// ★TCP受信
	BOOL RecvUDPData(pollfd *fds);				// ★UDP受信
	BOOL SendTCPData(pollfd *fds, DWORD dwNow);	// ★TCP送信
	BOOL SendUDPData(pollfd *fds, DWORD dwNow);	// ★UDP送信

	int AnalyzeDataRecv(int iTcpUdp);							// ★受信データの解析(TCP/UDPで参照バッファ切り替え)
	int RecvMessagePacket(HeaderRec *pHeader, int iTcpUdp);		// ★メッセージ受信(TCP/UDPで参照バッファ切り替え)
	int RecvFileSendPacket(HeaderRec *pHeader, int iTcpUdp);	// ★ファイル送信コマンド受信関数(TCP/UDPで参照バッファ切り替え)
	int RecvFileRecvPacket(HeaderRec *pHeader, int iTcpUdp);	// ★ファイル受信コマンド受信関数(TCP/UDPで参照バッファ切り替え)
	int RecvTransMethodPacket(HeaderRec *pHeader);				// ★データ送受信方法受信関数(TCPのみ)
	BOOL MakePeerUDPSockaddr(WORD wPeerPort);					// ★UDP通信用に相手先sockaddrを作る(ポート番号付与)

	DWORD CalcNextSendInterval(int iSentSize);	// 次回送信までの間隔
	BOOL CanSendNow(DWORD dwNow, int iTcpUdp);	// ★送信して良い時刻になったか(TCP/UDPで参照変数切り替え)

	BOOL KillZombei();							// このクラスが作った不要(エラー発生)になったスレッドを破棄
};

