#include	"stdThread.h"
#include	"SendRecvThread.h"	// 送受信スレッドを使うため
#include	"MySyncObject.h"	// 同期オブジェクトを使うため
#include	"define.h"

#define	MAXPENDING			(5)		// 接続バックログ数(保留中の接続のキューの最大長)
#define MAX_SOCKET_NUM		(20)	// 接続待用ソケットの最大数
#define MAX_CONNECTION_NUM	(10)	// 最大受け入れ接続数

// 関数の宣言
BOOL CreateAndBindSocket(WORD wPort);	// 接続待ち用ソケットの作成と名前付け
BOOL DestroySocket(SOCKET &fd);			// 切断とソケットの破棄		
BOOL DoListen();						// ソケットを接続待ちにする
BOOL DoAccept(WORD wStartUDPSrcPort);	// ★接続の受容(接続済みソケットを作成する)UDPの送信元ポート番号を指定可に
void Stop();							// すべてのソケットを破棄する
BOOL CheckKey();						// キー入力の検査
BOOL KillZombei();						// ゾンビ状態のスレッドを破棄する
BOOL SerializeSendData();				// データをシリアライズし各送受信スレッドの送信バッファに格納する

// 変数の宣言
SOCKET	m_fdServer[MAX_SOCKET_NUM];		// 待機用ソケット
int		m_iSockCount;					// 待機中ソケットの数
CSendRecvThread		*m_pCSendRecvThread[MAX_CONNECTION_NUM] = { NULL };
ConnectionInfoRec	*m_pConInfo[MAX_CONNECTION_NUM] = { NULL };
CMySyncObject		*m_pCMySyncObject = NULL;

int main(int argc, char *argv[])
{
	int		iRet = -1;
	int		ii;
	WORD	wStartUDPSrcPort = 0;		// ★UDPの送信元ポート番号を指定可に

	EnableESC();
	Locate(1, 1, 2);
	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();
	// ★起動パラメータチェック UDP送信元開始ポート番号を追加
	if ((argc != 2) && (argc != 3))
	{
		fprintf(stderr, "Usage:  %s <Server Port> [Start UDP Src Port]\n", argv[0]);
		goto L_END;
	}
	// ★UDP送信元ポート 指定がないときはbindで決定される動的ポートを使用する
	if (argc == 3)
		wStartUDPSrcPort = (WORD)atol(argv[2]);
	// 変数の初期化
	for (ii = 0; ii < MAX_SOCKET_NUM; ++ii)
		m_fdServer[ii] = INVALID_SOCKET;

	// 接続待ちソケットの作成と名前付け
	// bind済みのソケットがm_fdServer[]に格納される
	if (CreateAndBindSocket((WORD)atol(argv[1])) == FALSE)
		goto L_END;
	// 接続待機中状態にする
	// m_fdServer[]に格納されているソケットが接続待ちになる
	if (DoListen() == FALSE)
		goto L_END;
	// 接続要求に対して接続済みソケットを作成する
	// 送受信スレッドの作成をゾンビスレッドの破棄を行う
	// ★送受信スレッド作成時にUDP送信元ポートを渡す
	if (DoAccept(wStartUDPSrcPort) == FALSE)
		goto L_END;

	iRet = 0;
L_END:
	// 切断とすべてのソケットの破棄
	Stop();
	m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	return(iRet);
}

//----------------------------------------------
// function
//    待機用ソケットの作成とbindの実施
// parameter
//    WORD	wPort	[in]ポート番号
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL CreateAndBindSocket(WORD wPort)
{
	BOOL	fRet = TRUE;
	struct addrinfo	hints, *pres = NULL, *pTemp = NULL;
	char	szPort[NI_MAXSERV];
	int 	on;

	fprintf(stderr, "CreateAndBindSocket()\n");

	m_iSockCount = 0;
	// ストリーム型で待機可能なアドレス情報の条件
	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	DISABLE_C4996
	sprintf(szPort, "%d", wPort);
	ENABLE_C4996

	if (getaddrinfo(NULL, szPort, &hints, &pres) == 0)
	{
		// 順番にpresの中に格納されている情報を使用する
		pTemp = pres;
		while ((pTemp != NULL) && (m_iSockCount < MAX_SOCKET_NUM))
		{
			// Socketの作成, bindの実行
			// ソケット作成
			m_fdServer[m_iSockCount] = socket(pTemp->ai_family, pTemp->ai_socktype, pTemp->ai_protocol);
			if (m_fdServer[m_iSockCount] == INVALID_SOCKET)
				goto L_NEXT;
			fprintf(stderr, "%d %d %d %d\n", (int)m_fdServer[m_iSockCount], pTemp->ai_family, AF_INET, AF_INET6);
#ifdef IPV6_V6ONLY
			// IPv6ソケットでIPv4射影アドレスを使用しないように設定
			if (pTemp->ai_family == AF_INET6)
			{
				on = 1;
				if (setsockopt(m_fdServer[m_iSockCount], IPPROTO_IPV6, IPV6_V6ONLY, (char *)&on, sizeof(on)) < 0)
				{
					DispErrorMsg("Err:setsockopt");
					goto L_NEXT;
				}
			}
#endif
			// クローズ直後に際bindできない状態の解消
			on = 1;
			setsockopt(m_fdServer[m_iSockCount], SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
			// bindの実行
			if (bind(m_fdServer[m_iSockCount], pTemp->ai_addr, (int)pTemp->ai_addrlen) == SOCKET_ERROR)
			{
				// bindに失敗したら、そのソケットを破棄して次のソケットの処理を行う
				DispErrorMsg("Err:bind");
				DestroySocket(m_fdServer[m_iSockCount]);
				goto L_NEXT;
			}
			++m_iSockCount;
		L_NEXT:
			pTemp = pTemp->ai_next;			// 次の情報を処理する
		}
		freeaddrinfo(pres);
	}
	if (m_iSockCount == 0)					// ひとつもbindできなかったときは失敗を返す
		fRet = FALSE;
	return(fRet);
}

//----------------------------------------------
// function
//    待機用ソケットを接続待ち状態にする
// parameter
//    なし
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DoListen()
{
	BOOL		fRet = FALSE;
	int			ii;

	fprintf(stderr, "DoListen()\n");
	for (ii = 0; ii < m_iSockCount; ++ii)
	{
		if (listen(m_fdServer[ii], MAXPENDING) == SOCKET_ERROR)
			continue;
		fRet = TRUE;
	}
	return(fRet);
}

//----------------------------------------------
// function
//    ★接続要求の受け入れ(接続済みソケットの作成)
//    送受信スレッドの管理
// parameter
//    WORD	wStartUDPSrcPort	[in]UDP送信元開始ポート番号/0
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DoAccept(WORD wStartUDPSrcPort)
{
	BOOL				fRet = FALSE;
	sockaddr_storage	ClntAddr;	// 接続クライアントのアドレス情報
	socklen_t			iClntLen;	// 接続クライアントのアドレス情報のサイズ
	char				szHostAddr[NI_MAXHOST];
	int					ii;
	pollfd				fds[MAX_SOCKET_NUM] = {0};
	int					jj;			// 最大接続数まで送受信スレッドを動かす
	SOCKET				fdClient;	// そのためにacceptの結果を記憶する

	fprintf(stderr, "DoAccept()\n");
	for (ii = 0; ii < m_iSockCount; ++ii)
	{
		fds[ii].fd = m_fdServer[ii];
		fds[ii].events = POLLIN;				// 受信イベントを設定
	}

	while (1)
	{
		KillZombei();				// ゾンビ状態のスレッドは破棄する
		SerializeSendData();
		poll(fds, m_iSockCount, 10);
		// キー入力で中断
		if (CheckKey())
		{
			fprintf(stderr, "  OK:Abort by key\n");
			break;
		}
		for (ii = 0; ii < m_iSockCount; ++ii)
		{
			if (fds[ii].revents & POLLERR)			// エラー発生
				DispErrorMsg("Err:DoAccept");
			else if(fds[ii].revents & POLLIN)		// 受信イベント発生
			{
				iClntLen = sizeof(ClntAddr);		// 接続先アドレス情報を格納する構造体のサイズ
				// 受容したときのソケットを記憶し、送受信スレッドを作るか切断するかに使用
				if ((fdClient = accept(m_fdServer[ii], (sockaddr *)&ClntAddr, &iClntLen)) == INVALID_SOCKET)
				{
					// 失敗時次の接続要求を処理する
					DispErrorMsg("Err:accept");
					continue;
				}
				else
				{
					// 成功時fdClientには接続済みのソケットが格納されている
					// 相手の情報から、IPアドレスを調べる
					if (getnameinfo((struct sockaddr *) &ClntAddr, (socklen_t)iClntLen,
						szHostAddr, sizeof(szHostAddr), NULL, 0, NI_NUMERICHOST) == 0)
					{
						fprintf(stderr, "%s\n", szHostAddr);
					}
					// 接続数に余裕があればCSendRecvThreadのインスタンスを作成し実行
					// なければ、切断する
					fRet = FALSE;
					for (jj = 0; jj < MAX_CONNECTION_NUM; ++jj)
					{
						if (m_pCSendRecvThread[jj] == NULL)
						{
							m_pConInfo[jj] = (ConnectionInfoRec *)calloc(1, sizeof(ConnectionInfoRec));
							m_pConInfo[jj]->pCMySyncObject = m_pCMySyncObject;
							m_pConInfo[jj]->fdClient = fdClient;
							if (wStartUDPSrcPort > 0)
								m_pConInfo[jj]->wUDPSrcPort = wStartUDPSrcPort + jj;	// ★指定のUDP送信元ポートを使用
							else
								m_pConInfo[jj]->wUDPSrcPort = 0;						// ★動的UDP送信元ポートを使用
							m_pConInfo[jj]->ai_addr = (sockaddr *)&ClntAddr;			// ★UDP送信のために相手アドレスを渡す
							m_pConInfo[jj]->ai_addrlen = iClntLen;

							m_pCSendRecvThread[jj] = new CSendRecvThread(m_pConInfo[jj]);
							m_pCSendRecvThread[jj]->Begin();
							fRet = TRUE;
							break;
						}
					}
					if (fRet == FALSE)
						DestroySocket(fdClient);
				}
			}
		}
	}
	return(fRet);
}


//----------------------------------------------
// function
//    TCPソケットの破棄
// parameter
//    SOCKET		&fd[in/out]破棄するソケット
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DestroySocket(SOCKET &fd)
{
	if (fd != INVALID_SOCKET)
	{
		fprintf(stderr, "DestroySocket()\n");
		shutdown(fd, SHUT_RDWR);     // 受信も送信も停止
		close(fd);
		fd = INVALID_SOCKET;
	}
	return(TRUE);
}

//----------------------------------------------
// function
//    切断、ソケットの破棄、受信スレッドの破棄
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void Stop()
{
	int		ii;

	fprintf(stderr, "Stop()\n");
	//受信スレッドの破棄, 接続済みソケットの破棄
	for (ii = 0; ii < MAX_CONNECTION_NUM; ++ii)
	{
		if (m_pCSendRecvThread[ii] != NULL)
		{
			m_pCSendRecvThread[ii]->End();
			m_pCSendRecvThread[ii]->WaitForEnd();
			SAFE_DELETE(m_pCSendRecvThread[ii]);

			DestroySocket(m_pConInfo[ii]->fdClient);
			SAFE_FREE(m_pConInfo[ii])
		}
	}
	// 接続待ちソケットの破棄
	for (ii = 0; ii < m_iSockCount; ++ii)
		DestroySocket(m_fdServer[ii]);
}

//----------------------------------------------
// function
//	  キー入力チェック
//	  標準入力は0なのでselectのままでOKですが、練習のためにpollに
// parameter
//    なし
// return
//    TRUE:入力あり
//----------------------------------------------
BOOL CheckKey()
{
	BOOL			fRet = FALSE;
	pollfd			fds[1] = { 0 };
	char			szBuff[81];

	fds[0].fd = 0;					// 0:標準入力 1:標準出力 2:標準エラー
	fds[0].events = POLLIN;			// Enterキー入力を対象に
	poll(fds, 1, 10);
	if (fds[0].revents & POLLIN)	// ENTERキー入力があった
	{
		read(0, szBuff, sizeof(szBuff));
		fRet = TRUE;
	}
	return(fRet);
}

//----------------------------------------------
// function
//	切断済みのCSendRecvThreadを破棄する
// parameter
//	なし
// return
//	TRUE:破棄した/FALSE:破棄すべきものがなかった
//----------------------------------------------
BOOL KillZombei()
{
	BOOL	fRet = FALSE;
	int		ii;

	for (ii = 0; ii < MAX_CONNECTION_NUM; ++ii)
	{
		if (m_pCSendRecvThread[ii] != NULL)
		{
			if (m_pCSendRecvThread[ii]->IsZombie() == TRUE)
			{
				m_pCSendRecvThread[ii]->End();
				m_pCSendRecvThread[ii]->WaitForEnd();
				SAFE_DELETE(m_pCSendRecvThread[ii]);

				DestroySocket(m_pConInfo[ii]->fdClient);
				SAFE_FREE(m_pConInfo[ii])
					fRet = TRUE;
			}
		}
	}
	return(fRet);
}

//----------------------------------------------
// function
//	ゾンビ状態でないCSendRecvThreadに送信すべきデータがあれば
//	シリアライズして送信バッファに格納する
// parameter
//	なし
// return
//	TRUE:セットすべきデータがあった/FALSE:なかった
//----------------------------------------------
BOOL SerializeSendData()
{
	BOOL	fRet = FALSE;
	int		ii, iRecvSize = 0, iRecvIndex = -1;
	char	*pcRecvData = NULL;

	// チャットデータの分配(パケット単位で取り出し分配)
	// ゾンビ状態でない受信したClient以外のすべてに送信データとしてセットする
	for (ii = 0; ii < MAX_CONNECTION_NUM; ++ii)
	{
		if (m_pCSendRecvThread[ii] != NULL)
		{
			if (m_pCSendRecvThread[ii]->IsZombie() == FALSE)
			{
				iRecvSize = m_pCSendRecvThread[ii]->GetDataForDistributeChat(&pcRecvData);
				if (iRecvSize > 0)
				{
					fRet = TRUE;
					iRecvIndex = ii;
					break;
				}
			}
		}
	}
	if (fRet == TRUE)
	{
		for (ii = 0; ii < MAX_CONNECTION_NUM; ++ii)
		{
			if ((m_pCSendRecvThread[ii] != NULL) && (ii != iRecvIndex))
			{
				if (m_pCSendRecvThread[ii]->IsZombie() == FALSE)
				{
					m_pCSendRecvThread[ii]->SetSendData(pcRecvData, iRecvSize);
				}
			}
		}
	}
	SAFE_FREE(pcRecvData)
	return(fRet);
}

